// Copyright (c) 2006, TopCoder, Inc. All rights reserved.
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Xml;
using TopCoder.Uml.Xmi.MetaData;
using TopCoder.Uml.Xmi.Parser;
using TopCoder.Util.File;
using TopCoder.Util.File.TemplateSource;

namespace TopCoder.Uml.Xmi
{
    /// <summary>
    /// The main class for the code generator.  Handes using the XMI parser and document generator to
    /// generate source code.
    /// </summary>
    /// <author>Dan Manges (Seiz3r)</author>
    /// <copyright>Copyright (c) 2006, TopCoder, Inc. All rights reserved.</copyright>
    public class CodeGenerator
    {
        /// <summary>
        /// The code template file.
        /// </summary>
        private const string TEMPLATE_FILE = @"..\..\conf\code_template.txt";

        /// <summary>
        /// The source XMI file.
        /// </summary>
        private string source;

        /// <summary>
        /// The destination directory to create source code files.
        /// </summary>
        private string destination;

        /// <summary>
        /// The main entry point for the program.  Expects two arguments, the source XMI file and the
        /// destination directory.  Unless those two arguments are given, display usage statistics.
        /// </summary>
        /// <param name="args">The program arguments.</param>
        static void Main(string[] args)
        {
            if (args.Length != 2)
            {
                DisplayUsage();
            }
            else if (!File.Exists(args[0]))
            {
                Console.WriteLine("File could not be found.");
            }
            else if (!Directory.Exists(args[1]))
            {
                Console.WriteLine("Target directory does not exist.");
            }
            else
            {
                try
                {
                    CodeGenerator codeGenerator = new CodeGenerator(args[0], args[1]);
                    codeGenerator.Execute();
                    DisplaySuccessMessage();
                }
                catch (Exception ex)
                {
                    DisplayException(ex);
                }
            }
        }

        /// <summary>
        /// Write usage instructions to the console.
        /// </summary>
        static void DisplayUsage()
        {
            Console.WriteLine("Generates C# source code from an XMI file.");
            Console.WriteLine();
            Console.WriteLine("Usage: xmi2cs source destination");
            Console.WriteLine();
            Console.WriteLine("  source       The source XMI file.");
            Console.WriteLine("  destination  The target directory to place the source code.");
            Console.WriteLine();
            Console.WriteLine(@"For TopCoder components, the recommended destination is src\csharp\main");
        }

        /// <summary>
        /// Write exception information to the console.
        /// </summary>
        /// <param name="ex">The exception to display.</param>
        static void DisplayException(Exception ex)
        {
            Console.WriteLine(ex.GetType().FullName);
            Console.WriteLine(ex.Message);
            Console.WriteLine(ex.StackTrace);
        }

        /// <summary>
        /// Write a code generation successful message to the console.
        /// </summary>
        static void DisplaySuccessMessage()
        {
            Console.WriteLine();
            Console.WriteLine("Code generation was successful.");
            Console.WriteLine();
            Console.WriteLine("After compiling the source code, run the Unit Test Generator on your assembly file " +
                "to generate test code stubs.  Type \"gen\" for command line options.");
        }

        /// <summary>
        /// Create a new instance of <c>CodeGenerator</c>, setting the source file and destination directory.
        /// </summary>
        /// <param name="source">The source XMI file.</param>
        /// <param name="destination">The destination directory.</param>
        public CodeGenerator(string source, string destination)
        {
            this.source = source;
            this.destination = destination;
        }

        /// <summary>
        /// Parse the XMI file and generate source code.
        /// </summary>
        public void Execute()
        {
            XmiHandler handler = new XmiHandler();
            XmiParser parser = new XmiParser();
            using (StreamReader reader = new StreamReader(this.source))
            {
                parser.AddHandler(XmiHandler.NAME, handler, XmiHandler.NODE_TYPES);
                parser.Parse(reader);             
            }
            foreach (UmlClass umlClass in handler.Classes)
            {
                if (umlClass.Namespace.Contains("TopCoder"))
                {
                    this.WriteClass(umlClass);
                }
            }
        }

        /// <summary>
        /// Write an XML file to be used with the document generator to generate the source code.
        /// </summary>
        /// <param name="umlClass">The class to write to XML.</param>
        private void WriteClass(UmlClass umlClass)
        {
            // The XML file to write to.
            string file = Helper.FileForClass(this.destination, umlClass.Namespace, umlClass.Name, "xml");
            using (Stream stream = File.OpenWrite(file))
            {
                using (XmlWriter writer = XmlWriter.Create(stream))
                {
                    // Pass the XmlWriter to the class to write.
                    umlClass.WriteXml(writer);
                }
            }
            // Generate the source code based on the XML file.
            this.GenerateCode(file);
            // Delete the XML file.
            File.Delete(file);
        }

        /// <summary>
        /// Generate csharp code based on XML data and a template file.
        /// </summary>
        /// <param name="xmlFile">The XML file from which to extract data.</param>
        private void GenerateCode(string xmlFile)
        {
            // The file to write.
            string outputFile = xmlFile.Replace(".xml", ".cs");
            // Read the contents of the template file into a string.
            string templateContent = "";
            using (StreamReader reader = new StreamReader(TEMPLATE_FILE))
            {
                templateContent = reader.ReadToEnd();
            }
            // Create a new instance of a DocumentGenerator and create a ITemplate to use.
            DocumentGenerator generator = DocumentGenerator.GetInstance();
            ITemplate template = generator.ParseTemplate(templateContent);
            // Create a reader for the XML file and a writer for the CS file.
            using (StreamReader xmlFileReader = new StreamReader(xmlFile))
            {
                using (StreamWriter csFileWriter = new StreamWriter(outputFile))
                {
                    // Apply the template to the source data and write the results to the csFileWriter.
                    generator.ApplyTemplate(template, xmlFileReader, csFileWriter);
                }
            }
            this.CleanSourceCodeFile(outputFile);
        }

        /// <summary>
        /// Cleans up newline characters in the source code file.
        /// </summary>
        /// <param name="file">The file to clean.</param>
        private void CleanSourceCodeFile(string file)
        {
            StringBuilder builder;
            using (StreamReader reader = new StreamReader(file))
            {
                builder = new StringBuilder(reader.ReadToEnd());
            }
            builder.Replace("\r\n", "\n");
            builder.Replace("\r", "\n");
            builder.Replace("\n", "\r\n");
            using (StreamWriter writer = new StreamWriter(file))
            {
                writer.Write(builder.ToString());
            }
        }
    }
}