// Copyright (c) 2006, TopCoder, Inc. All rights reserved.
using System;
using System.Collections.Generic;
using System.Reflection;
using System.Xml;
using TopCoder.Uml.Xmi.MetaData;
using TopCoder.Uml.Xmi.Parser;

namespace TopCoder.Uml.Xmi
{
    /// <summary>
    /// Implements <c>IXmiHandler</c> to collect data from an XMI file.  This class collects the majority of data
    /// necessary to generate source code.
    /// </summary>
    /// <author>Dan Manges (Seiz3r)</author>
    /// <copyright>Copyright (c) 2006, TopCoder, Inc. All rights reserved.</copyright>
    public class XmiHandler : IXmiHandler
    {
        /// <summary>
        /// Name to use to register handler with an <c>XmiParser</c>
        /// </summary>
        public const string NAME = "Code Generation Handler";

        /// <summary>
        /// Node types to collect in this handler, which are passed to the <c>XmiParser</c>
        /// </summary>
        public static string[] NODE_TYPES = { "UML:Package", "UML:Class", "UML:Operation",
            "UML:Parameter", "UML:DataType", "UML:Stereotype", "UML:Interface", "UML:Attribute",
            "UML:TagDefinition" };

        /// <summary>
        /// A dictionary of UML:Package IDs and their corresponding namespace.
        /// </summary>
        private Dictionary<string, string> packages = new Dictionary<string, string>();

        /// <summary>
        /// A dictionary of UML:Class nodes referenced by xmi.id
        /// </summary>
        private Dictionary<string, UmlClass> classes = new Dictionary<string, UmlClass>();

        /// <summary>
        /// A dictionary of methods, or UML:Operation nodes referenced by xmi.id
        /// </summary>
        private Dictionary<string, Method> methods = new Dictionary<string, Method>();

        /// <summary>
        /// A dictionary of method parameters, referenced by xmi.id
        /// </summary>
        private Dictionary<string, UmlParameter> parameters = new Dictionary<string, UmlParameter>();

        /// <summary>
        /// A dictionary of UML:DataType nodes, referencing the type as a string by the xmi.id
        /// </summary>
        private Dictionary<string, string> dataTypes = new Dictionary<string, string>();

        /// <summary>
        /// A dictionary of stereotype strings referenced by xmi.id
        /// </summary>
        private Dictionary<string, string> stereotypes = new Dictionary<string, string>();

        /// <summary>
        /// Dictionary of UML:TagDefinition nodes, which store documentation.
        /// </summary>
        private Dictionary<string, string> tagDefinitions = new Dictionary<string, string>();

        /// <summary>
        /// Dictionary of UML:Attribute nodes, referenced by xmi.id
        /// </summary>
        private Dictionary<string, UmlAttribute> attributes = new Dictionary<string, UmlAttribute>();

        /// <summary>
        /// The last UML:Package id encountered by the parser.
        /// </summary>
        private string lastPackageId = "";

        /// <summary>
        /// The last UML:Class id encountered by the parser.
        /// </summary>
        private string lastClassId = "";

        /// <summary>
        /// The last UML:Operation id encountered by the parser.
        /// </summary>
        private string lastMethodId = "";
        
        /// <summary>
        /// Process a node passed from the parser.  This is the method the parser looks for from IXmiHandler.
        /// </summary>
        /// <remarks>Uses reflection to pass the node to a corresponding private method.</remarks>
        /// <param name="node">The node passed from the parser.</param>
        /// <returns>A string representation of the node.</returns>
        public string ProcessNode(XmlNode node)
        {
            string method = "Process" + node.LocalName;
            this.GetType().InvokeMember(method,
              BindingFlags.InvokeMethod | BindingFlags.NonPublic | BindingFlags.Instance,
              null, this, new object[] { node });
            return node.Name;
        }

        /// <summary>
        /// After the entire XMI file has been parsed, call this method to update data.
        /// </summary>
        /// <remarks>
        /// Because nodes may be referenced before they are encountered, many objects must store id references and
        /// then make updates when parsing is complete.
        /// </remarks>
        public void EndProcessing()
        {
            this.SetDocumentationTypes();
            this.SetParameterTypes();
            this.SetAttributeTypes();
            this.MapStereotypes();
            this.MapParametersToMethods();
            this.MapAttributesToClasses();
            this.MapMethodsToClasses();
        }

        /// <summary>
        /// Replace stereotype references with stereotype strings in methods and attributes.
        /// </summary>
        private void MapStereotypes()
        {
            foreach (Method m in this.methods.Values)
            {
                foreach (string idRef in m.StereotypeReferences)
                {
                    this.AddStereotype(m, idRef);
                }
            }
            foreach (UmlAttribute a in this.attributes.Values)
            {
                foreach (string idRef in a.StereotypeReferences)
                {
                    this.AddStereotype(a, idRef);
                }
            }
        }

        /// <summary>
        /// Look up a stereotype from an id reference and add it to the XmiElement.
        /// </summary>
        /// <param name="xmiElement">The XmiElement to add the stereotype.</param>
        /// <param name="stereotypeIdRef">The stereotype id to look up.</param>
        private void AddStereotype(XmiElement xmiElement, string stereotypeIdRef)
        {
            if (this.stereotypes.ContainsKey(stereotypeIdRef))
            {
                xmiElement.AddStereotype(this.stereotypes[stereotypeIdRef]);
            }
        }

        /// <summary>
        /// Process a UML:Attribute from the XmiParser.
        /// </summary>
        /// <param name="node">The node from the parser.</param>
        private void ProcessAttribute(XmlNode node)
        {
            // Grab the id, return if it is no present.
            string xmiId = Helper.ReadAttribute(node, "xmi.id");
            if (xmiId == null)
            {
                return;
            }
            // Store the attribute in the lookup table and add a reference to the most recent class.
            this.attributes[xmiId] = new UmlAttribute(node);
            this.classes[lastClassId].AddAttributeReference(xmiId);
        }

        /// <summary>
        /// Process a UML:DataType node from the XmiParser.
        /// </summary>
        /// <param name="node">The node from the parser.</param>
        private void ProcessDataType(XmlNode node)
        {
            string xmiId = Helper.ReadAttribute(node, "xmi.id");
            string name = Helper.ReadAttribute(node, "name");
            // Store a reference to the data type if the id and name are present.
            if (xmiId != null && name != null)
            {
                this.dataTypes.Add(xmiId, name);
            }
        }

        /// <summary>
        /// Process a UML:TagDefinition node from the XmiParser.
        /// </summary>
        /// <param name="node">The node from the parser.</param>
        private void ProcessTagDefinition(XmlNode node)
        {
            string xmiId = Helper.ReadAttribute(node, "xmi.id");
            string name = Helper.ReadAttribute(node, "name");
            if (xmiId != null && name != null)
            {
                this.tagDefinitions.Add(xmiId, name);
            }
        }

        /// <summary>
        /// Process a UML:Stereotype node from the XmiParser.
        /// </summary>
        /// <param name="node">The node from the parser.</param>
        private void ProcessStereotype(XmlNode node)
        {
            string xmiId = Helper.ReadAttribute(node, "xmi.id");
            string name = Helper.ReadAttribute(node, "name");
            if (xmiId != null && name != null)
            {
                this.stereotypes.Add(xmiId, name);
            }
        }

        /// <summary>
        /// Process a UML:Package node from the XmiParser.
        /// </summary>
        /// <param name="node">The node from the parser.</param>
        private void ProcessPackage(XmlNode node)
        {
            string xmiId = Helper.ReadAttribute(node, "xmi.id");
            if (xmiId == null)
            {
                return;
            }
            this.lastPackageId = xmiId;
            this.packages[xmiId] = Helper.ReadAttribute(node, "name");
        }

        /// <summary>
        /// Process a UML:Interface node from the XmiParser.
        /// </summary>
        /// <param name="node">The node from the parser.</param>
        private void ProcessInterface(XmlNode node)
        {
            string xmiId = Helper.ReadAttribute(node, "xmi.id");
            if (xmiId == null)
            {
                return;
            }
            // Store the interface in with the classes.
            this.lastClassId = xmiId;
            this.classes[xmiId] = new UmlClass(node, true);
            // Set the namespace to the most recent package.
            this.classes[xmiId].Namespace = this.packages[lastPackageId];
        }

        /// <summary>
        /// Process a UML:Class node from the XmiParser.
        /// </summary>
        /// <param name="node">The node from the parser.</param>
        private void ProcessClass(XmlNode node)
        {
            string xmiId = Helper.ReadAttribute(node, "xmi.id");
            if (xmiId == null)
            {
                return;
            }
            if (node.ChildNodes.Count > 0)
            {
                this.lastClassId = xmiId;
            }
            this.classes[xmiId] = new UmlClass(node);
            this.classes[xmiId].Namespace = this.packages[lastPackageId];
        }

        /// <summary>
        /// Process a UML:Operation node from the XmiParser.
        /// </summary>
        /// <param name="node">The node from the parser.</param>
        private void ProcessOperation(XmlNode node)
        {
            string xmiId = Helper.ReadAttribute(node, "xmi.id");
            if (xmiId == null)
            {
                return;
            }
            this.lastMethodId = xmiId;
            this.methods[xmiId] = new Method(node);
            this.classes[lastClassId].AddMethodReference(xmiId);
        }

        /// <summary>
        /// Process a UML:Parameter node from the XmiParser.
        /// </summary>
        /// <param name="node">The node from the parser.</param>
        private void ProcessParameter(XmlNode node)
        {
            string xmiId = Helper.ReadAttribute(node, "xmi.id");
            if (xmiId == null)
            {
                return;
            }
            this.parameters[xmiId] = new UmlParameter(node);
            this.methods[lastMethodId].AddParameterReference(xmiId);
        }

        /// <summary>
        /// Get an array of UmlClass objects extracted from the XMI.
        /// </summary>
        /// <value>An array of UmlClass objects extracted from the XMI.</value>
        public UmlClass[] Classes
        {
            get
            {
                // Add the classes to a list then return as an array.
                List<UmlClass> list = new List<UmlClass>();
                foreach (UmlClass c in this.classes.Values)
                {
                    list.Add(c);
                }
                return list.ToArray();
            }
        }

        /// <summary>
        /// After parsing is complete, set the documentation types for documentation nodes.
        /// </summary>
        /// <remarks>Loops through each class, method, attribute, and parameter to set documentation.</remarks>
        private void SetDocumentationTypes()
        {
            foreach (XmiElement element in this.classes.Values)
            {
                SetDocumentation(element);
            }
            foreach (XmiElement element in this.methods.Values)
            {
                SetDocumentation(element);
            }
            foreach (XmiElement element in this.attributes.Values)
            {
                SetDocumentation(element);
            }
            foreach (XmiElement element in this.parameters.Values)
            {
                SetDocumentation(element);
            }
        }

        /// <summary>
        /// Read each UML:TaggedValue in the XmiElement and set a DocumentationType for the documentation node.
        /// </summary>
        /// <param name="xmiElement">The element to set the documentation type.</param>
        private void SetDocumentation(XmiElement xmiElement)
        {
            foreach (UmlTaggedValue taggedValue in xmiElement.UmlTaggedValues)
            {
                taggedValue.DocumentationType = Helper.StringToDocumentationType(this.tagDefinitions[taggedValue.TagDefinitionIdRef]);
            }
        }

        /// <summary>
        /// Set the type of attributes after parsing the entire XMI document.
        /// </summary>
        /// <remarks>
        /// The type could reference a class, interface, or datatype.  This method simplifies getting a type name from
        /// any of the three.
        /// </remarks>
        private void SetAttributeTypes()
        {
            foreach (UmlAttribute umlAttribute in this.attributes.Values)
            {
                umlAttribute.Type = this.GetTypeObjectName(umlAttribute.TypeRef.UmlType, umlAttribute.TypeRef.IdRef);
                umlAttribute.Namespace = this.GetTypeNamespace(umlAttribute.TypeRef.UmlType, umlAttribute.TypeRef.IdRef);
            }
        }

        /// <summary>
        /// Set the type of parameters after parsing the entire XMI document.
        /// </summary>
        /// <reamrks>
        /// The typte could reference a class, interface, or datatype.
        /// </reamrks>
        private void SetParameterTypes()
        {
            foreach (UmlParameter p in this.parameters.Values)
            {
                p.Type = this.GetTypeObjectName(p.TypeRef.UmlType, p.TypeRef.IdRef);
                // Set the namespace to help with setting the include statements.
                p.Namespace = this.GetTypeNamespace(p.TypeRef.UmlType, p.TypeRef.IdRef);
            }
        }

        /// <summary>
        /// After parsing is complete, map classes' attribute references to the actual attribute instances.
        /// </summary>
        /// <remarks>
        /// This is necessary because a class could reference a UML:Attribute node before the parser has encountered it.
        /// </remarks>
        private void MapAttributesToClasses()
        {
            foreach (UmlClass @class in this.classes.Values)
            {
                foreach (string attributeIdRef in @class.AttributeReferences)
                {
                    if (attributes.ContainsKey(attributeIdRef))
                    {
                        UmlAttribute a = attributes[attributeIdRef];
                        @class.AddUmlAttribute(a);
                    }
                }
            }
        }

        /// <summary>
        /// After parsing is complete, map classes' method references to the actual method instances.
        /// </summary>
        /// <remarks>
        /// This is necessary because a class could reference a UML:Operating node before the parser has encountered it.
        /// </remarks>
        private void MapMethodsToClasses()
        {
            foreach (UmlClass c in this.classes.Values)
            {
                foreach (string methodIdRef in c.MethodReferences)
                {
                    if (methods.ContainsKey(methodIdRef))
                    {
                        Method m = methods[methodIdRef];
                        c.AddMethod(m);
                    }
                }
            }
        }

        /// <summary>
        /// After parsing is complete, map methods' parameter references to the actual parameter instance.
        /// </summary>
        /// <remarks>
        /// This is necessary because a method could reference a parameter node before the parser has encountered it.
        /// </remarks>
        private void MapParametersToMethods()
        {
            foreach (Method m in this.methods.Values)
            {
                foreach (string refId in m.ParameterReferences)
                {
                    if (parameters.ContainsKey(refId))
                    {
                        UmlParameter p = parameters[refId];
                        m.AddParameter(p);
                    }
                }
            }
        }

        /// <summary>
        /// Get the namespace for an UmlType, or default to "System"
        /// </summary>
        /// <param name="type">The type to use to search for the namespace.</param>
        /// <param name="id">The id of the object to look for.</param>
        /// <returns>The namespace.</returns>
        private string GetTypeNamespace(UmlType type, string id)
        {
            if (type == UmlType.Class)
            {
                return this.classes[id].Namespace;
            }
            return "System";
        }

        /// <summary>
        /// Get the type for a UmlType object, defaults to "object"
        /// </summary>
        /// <param name="type">The type to look for.</param>
        /// <param name="id">The id of the object to look for.</param>
        /// <returns>The type of the object.</returns>
        private string GetTypeObjectName(UmlType type, string id)
        {
            if (id == null)
            {
                return "object";
            }
            switch (type)
            {
                case UmlType.Class:
                case UmlType.Interface:
                    return this.classes[id].Name;
                case UmlType.DataType:
                    return this.dataTypes[id];
                default:
                    return "object";
            }
        }
    }
}